class TestCreation:
    def test_sg_creation(self, sg):
        pass  # fixture created SWIFTGalaxy

    def test_vr_creation(self, vr):
        pass  # fixture created Velociraptor interface

    def test_sg_vr_creation(self, sg_vr):
        pass  # fixture created SWIFTGalaxy with Velociraptor interface
